/*



*/

/**
   This program parses to an Abstract Syntax Tree (AST) strings
   from the language defined by the following grammar.

       Prog ::= Exp
              | '(' 'prog' (Fun | Exp)* Exp ')'

        Fun ::= '(' 'fun' VARIABLE Lambda ')'   // a function declaration

     Lambda ::= '(' 'lambda' VARIABLE* Exp ')'  // formal parameters followed by function body

        Exp ::= Apply
              | If
              | While
              | Var
              | Set
              | Begin
              | Print
              | BExp
              | AExp
              | INTEGER
              | BOOLEAN
              | VARIABLE

      Apply ::= '(' 'apply' Exp Exp* ')'  // function value followed by actual parameters

         If ::= '(' 'if' Exp Exp Exp ')'

      While ::= '(' 'while' Exp Exp ')'

        Var ::= '(' 'var' VARIABLE Exp ')'

        Set ::= '(' 'set' VARIABLE Exp ')'

      Begin ::= '(' 'begin' Exp* Exp ')'

      Print ::= '(' 'print' Exp ')'

       BExp ::= '(' '||'  Exp Exp+ ')'
              | '(' '&&'  Exp Exp+ ')'
              | '(' '!'   Exp ')'
              | '(' RelOp Exp Exp ')'
              | '('  EqOp Exp Exp ')'

      RelOp ::= '<' | '>' | '<=' | '>='
       EqOp ::= '==' | '!='

       AExp ::= '(' '+' Exp Exp* ')'
              | '(' '-' Exp Exp? ')'
              | '(' '*' Exp Exp+ ')'
              | '(' '/' Exp Exp  ')'
              | '(' '%' Exp Exp  ')'
              | '(' '^' Exp Exp  ')'

    INTEGER ::= [-][0-9]+
    BOOLEAN ::= 'true' | 'false'
   VARIABLE ::= [a-zA-Z][a-zA-Z0-9]*
*/

public class Parse2AST
{
   public static Tree parse(String expression) throws ParseException, TokenizeException
   {
      Tokenizer tokens = new Tokenizer(expression);

      return parse(tokens);
   }

   public static Tree parse(Tokenizer tokens) throws ParseException
   {
      if ( ! tokens.hasToken() )
         throw new ParseException("empty program\n");

      Tree result = getProg(tokens); // parse the tokens

      if ( tokens.hasToken() )  // there shouldn't be any more tokens
         throw new ParseException("unexpected input: "
                                   + tokens.peekToken().lexeme
                                   + " at line " + tokens.peekToken().line
                                   + ", position " + tokens.peekToken().position
                                   //+ "\n" + tokens
                                   + "\n");
      return result;
   }


   // Parse a prog.
   private static Tree getProg(Tokenizer tokens) throws ParseException
   {
      Tree result = null;

      // Check which kind of Prog we have.
      if ( tokens.peekToken().lexeme.equals("(")
        && tokens.has2Tokens()
        && tokens.peek2Token().lexeme.equals("prog") )
      {
         tokens.match("(");
         result = new Tree( tokens.nextToken() ); // "prog"
         boolean lastItemIsExp = false;

         if ( ! tokens.hasToken() )
            throw new ParseException("unexpected end of input while parsing prog\n");

         while ( ! tokens.peekToken().lexeme.equals(")") )
         {
            Tree temp = getFunOrExp(tokens);
            result.addSubTree(temp);
            lastItemIsExp = ! temp.getElement().lexeme.equals("fun");

            if ( ! tokens.hasToken() )
               throw new ParseException("unexpected end of input while parsing prog\n");
         }

         tokens.match(")"); // we know that there is a ")" token

         if (! lastItemIsExp)
            throw new ParseException("prog must end with an expression\n");
      }
      else // Parse the single expression.
      {
         result = getExp(tokens);
      }

      return result;
   }//getProg()


   // This function is needed by the REPL, so we make it public.
   public static Tree getFunOrExp(Tokenizer tokens) throws ParseException
   {
      if ( tokens.peekToken().lexeme.equals("(")
        && tokens.has2Tokens()
        && tokens.peek2Token().lexeme.equals("fun") )
      {
         return getFun(tokens);
      }
      else
      {
         return getExp(tokens);
      }
   }//getFunOrExp()


   // Parse a function definition.
   private static Tree getFun(Tokenizer tokens) throws ParseException
   {
      Tree result = null;


      return result;
   }//getFun()


   // Parse a lambda expression.
   private static Tree getLambda(Tokenizer tokens) throws ParseException
   {
      if ( ! tokens.peekToken().lexeme.equals("(")
        || ! tokens.has2Tokens()
        || ! tokens.peek2Token().lexeme.equals("lambda") )
      {
         throw new ParseException("expected lambda expression, found "
                                   + (tokens.has2Tokens()? tokens.peekToken().lexeme+tokens.peek2Token().lexeme
                                                         : tokens.peekToken().lexeme)
                                   + " at line " + tokens.peekToken().line
                                   + ", position " + tokens.peekToken().position
                                   //+ "\n" + tokens
                                   + "\n");
      }

      tokens.match("(");
      Tree result = new Tree( tokens.nextToken() ); // "lambda"

      if ( ! tokens.hasToken() )
         throw new ParseException("unexpected end of input while parsing lambda\n");

      int count = 0;

      // get the variable names
      while ( ! tokens.peekToken().lexeme.equals(")")
           && ! tokens.peekToken().lexeme.equals("(")
           && ! tokens.peekToken().lexeme.equals("true")
           && ! tokens.peekToken().lexeme.equals("false")
           && ! Character.isDigit(tokens.peekToken().lexeme.charAt(0)) )
      {
         result.addSubTree( getVariable(tokens) );
         count++;

         if ( ! tokens.hasToken() )
            throw new ParseException("unexpected end of input while parsing lambda\n");
      }

      // get the final expressions (if it wasn't one of the variable names)
      if (! tokens.peekToken().lexeme.equals(")"))
      {
         result.addSubTree( getExp(tokens) );
         count++;
      }

      // there must be at least the final expression
      if (0 == count)
         throw new ParseException("invalid lambda expression"
                                + " at line " + tokens.peekToken().line
                                + ", position " + tokens.peekToken().position
                                //+ "\n" + tokens
                                + "\n");

      if ( ! tokens.hasToken() )
         throw new ParseException("unexpected end of input while parsing lambda\n");

      if ( ! tokens.match(")") )
         throw new ParseException("expected \")\", found "
                                + tokens.peekToken().lexeme
                                + " at line " + tokens.peekToken().line
                                + ", position " + tokens.peekToken().position
                                //+ "\n" + tokens
                                + "\n");
      return result;
   }//getLambda()


   // Parse an expression.
   private static Tree getExp(Tokenizer tokens) throws ParseException
   {
      Tree result = null;

      if ( tokens.peekToken().lexeme.equals("(") )
      {
         if ( ! tokens.has2Tokens() )
         {
            throw new ParseException("unexpected end of input while parsing an expression\n");
         }

         if ( tokens.peek2Token().lexeme.equals("var") )
         {
            result = getVar(tokens);
         }
         else if ( tokens.peek2Token().lexeme.equals("set") )
         {
            result = getSet(tokens);
         }
       //else if ...

      }
    //else if ...



      return result;
   }//getExp()


   // Parse a function call.
   private static Tree getApply(Tokenizer tokens) throws ParseException
   {
      Tree result = null;


      return result;
   }//getApply()


   // Parse an if-expression.
   private static Tree getIf(Tokenizer tokens) throws ParseException
   {
      Tree result = null;


      return result;
   }//getIf()


   // Parse a while-loop expression.
   private static Tree getWhile(Tokenizer tokens) throws ParseException
   {
      Tree result = null;


      return result;
   }//getWhile()


   // Parse a var expression.
   private static Tree getVar(Tokenizer tokens) throws ParseException
   {
      Tree result = null;


      return result;
   }//getVar()


   // Parse a set expression.
   private static Tree getSet(Tokenizer tokens) throws ParseException
   {
      Tree result = null;


      return result;
   }//getSet()


   // Parse a begin expression.
   private static Tree getBegin(Tokenizer tokens) throws ParseException
   {
      Tree result = null;


      return result;
   }//getBegin()


   // Parse a print expression.
   private static Tree getPrint(Tokenizer tokens) throws ParseException
   {
      Tree result = null;


      return result;
   }//getPrint()


   // Parse a boolean expression.
   private static Tree getBexp(Tokenizer tokens) throws ParseException
   {
      Tree result = null;


      return result;
   }//getBexp()


   // Parse a relational expression (which is a kind of boolean expression).
   private static Tree getRexp(Tokenizer tokens) throws ParseException
   {
      Tree result = null;


      return result;
   }//getRexp()


   // Parse an arithmetic expression.
   private static Tree getAexp(Tokenizer tokens) throws ParseException
   {
      Tree result = null;


      return result;
   }//getAexp()

   private static Tree getVariable(Tokenizer tokens) throws ParseException
   {
      Tree result = null;


      return result;
   }//getVariable()
}//Parse2AST
